@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2004, Nate Nielsen
+ * Copyright (c) 2004, Stefan Walter
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without 
@@ -32,12 +32,14 @@
  * 
  *
  * CONTRIBUTORS
- *  Nate Nielsen <nielsen@memberwebs.com>
+ *  Stef Walter <stef@memberwebs.com>
  *  Andreas Steinmetz <ast@domdv.de>
  *  Rubio Vaughan <rubio@passim.net>
  *  Olivier Beyssac <ob@r14.freenix.org>
  */ 
 
+#define _GNU_SOURCE
+
 #include <sys/time.h>
 #include <sys/types.h>
 #include <sys/socket.h>
@@ -85,8 +87,8 @@ spthread_t;
 #define CRLF                "\r\n"
 
 #define SMTP_TOOLONG        "500 Line too long" CRLF
-#define SMTP_STARTBUSY      "554 Server Busy" CRLF
-#define SMTP_STARTFAILED    "554 Local Error" CRLF
+#define SMTP_STARTBUSY      "421 Server busy, too many connections" CRLF
+#define SMTP_STARTFAILED    "421 Local Error, cannot start thread" CRLF
 #define SMTP_DATAINTERMED   "354 Start mail input; end with <CRLF>.<CRLF>" CRLF
 #define SMTP_FAILED         "451 Local Error" CRLF
 #define SMTP_NOTSUPP        "502 Command not implemented" CRLF
@@ -110,6 +112,7 @@ spthread_t;
 #define ESMTP_BINARY        "BINARYMIME"
 #define ESMTP_CHECK         "CHECKPOINT"
 #define ESMTP_XCLIENT       "XCLIENT"
+#define ESMTP_XEXCH50       "XEXCH50" 
 
 #define HELO_CMD            "HELO"
 #define EHLO_CMD            "EHLO"
@@ -120,6 +123,7 @@ spthread_t;
 #define STARTTLS_CMD        "STARTTLS"
 #define BDAT_CMD            "BDAT"
 #define XCLIENT_CMD         "XCLIENT"
+#define XFORWARD_CMD        "XFORWARD"
 
 #define DATA_END_SIG        "." CRLF
 
@@ -158,7 +162,6 @@ spthread_t;
 #define CFG_TIMEOUT         "TimeOut"
 #define CFG_OUTADDR         "OutAddress"
 #define CFG_LISTENADDR      "Listen"
-#define CFG_HEADER      	"Header"
 #define CFG_TRANSPARENT     "TransparentProxy"
 #define CFG_DIRECTORY       "TempDirectory"
 #define CFG_KEEPALIVES      "KeepAlives"
@@ -199,6 +202,7 @@ static int make_connections(spctx_t* ctx, int client);
 static int read_server_response(spctx_t* ctx);
 static int parse_config_file(const char* configfile);
 static char* parse_address(char* line);
+static char* parse_xforward(char* line, const char* part);
 static const char* get_successful_rsp(const char* line, int* cont);
 static void do_server_noop(spctx_t* ctx);
 
@@ -275,28 +279,6 @@ int sp_run(const char* configfile, const char* pidfile, int dbg_level)
 
     sp_messagex(NULL, LOG_DEBUG, "starting up (%s)...", VERSION);
 
-    /* Create the socket */
-    sock = socket(SANY_TYPE(g_state.listenaddr), SOCK_STREAM, 0);
-    if(sock < 0)
-        err(1, "couldn't open socket");
-
-    fcntl(sock, F_SETFD, fcntl(sock, F_GETFD, 0) | FD_CLOEXEC);    
-    setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (void *)&true, sizeof(true));
-    
-    /* Unlink the socket file if it exists */
-    if(SANY_TYPE(g_state.listenaddr) == AF_UNIX)
-        unlink(g_state.listenname);
-
-    if(bind(sock, &SANY_ADDR(g_state.listenaddr), SANY_LEN(g_state.listenaddr)) != 0)
-        err(1, "couldn't bind to address: %s", g_state.listenname);
-
-    /* Let 5 connections queue up */
-    if(listen(sock, 5) != 0)
-    {
-        sp_message(NULL, LOG_CRIT, "couldn't listen on socket");
-        exit(1);
-    }
-      	
     /* Drop privileges before daemonizing */
     drop_privileges();
     
@@ -316,18 +298,46 @@ int sp_run(const char* configfile, const char* pidfile, int dbg_level)
         /* Open the system log */
         openlog(g_state.name, 0, LOG_MAIL);
     }
-
-    sp_messagex(NULL, LOG_DEBUG, "created socket: %s", g_state.listenname);
-
+    
     /* Handle some signals */
     signal(SIGPIPE, SIG_IGN); 
-    signal(SIGHUP,  SIG_IGN);
-    signal(SIGINT,  on_quit);
+    signal(SIGHUP, SIG_IGN);
+    signal(SIGINT, on_quit);
     signal(SIGTERM, on_quit);
 
     siginterrupt(SIGINT, 1);
     siginterrupt(SIGTERM, 1);
+    
+    /* Create the socket */
+    sock = socket(SANY_TYPE(g_state.listenaddr), SOCK_STREAM, 0);
+    if(sock < 0)
+    {
+        sp_message(NULL, LOG_CRIT, "couldn't open socket");
+        exit(1);
+    }
 
+    fcntl(sock, F_SETFD, fcntl(sock, F_GETFD, 0) | FD_CLOEXEC);    
+    setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (void *)&true, sizeof(true));
+    
+    /* Unlink the socket file if it exists */
+    if(SANY_TYPE(g_state.listenaddr) == AF_UNIX)
+        unlink(g_state.listenname);
+
+    if(bind(sock, &SANY_ADDR(g_state.listenaddr), SANY_LEN(g_state.listenaddr)) != 0)
+    {
+        sp_message(NULL, LOG_CRIT, "couldn't bind to address: %s", g_state.listenname);
+        exit(1);
+    }
+    
+    sp_messagex(NULL, LOG_DEBUG, "created socket: %s", g_state.listenname);
+    
+    /* Let 5 connections queue up */
+    if(listen(sock, 5) != 0)
+    {
+        sp_message(NULL, LOG_CRIT, "couldn't listen on socket");
+        exit(1);
+    }
+    
     pid_file(1);
 
     sp_messagex(NULL, LOG_DEBUG, "accepting connections");
@@ -542,7 +552,7 @@ static void connection_loop(int sock)
         /* Check to make sure we have a thread */
         if(fd != -1)
         {
-            sp_messagex(NULL, LOG_ERR, "too many connections open (max %d). sent 554 response", g_state.max_threads);
+            sp_messagex(NULL, LOG_ERR, "too many connections open (max %d). sent busy response", g_state.max_threads);
             write(fd, SMTP_STARTBUSY, KL(SMTP_STARTBUSY));
             shutdown(fd, SHUT_RDWR);
             close(fd);
@@ -569,6 +579,7 @@ static void connection_loop(int sock)
                 close(fd);
             }
           
+            sp_messagex(NULL, LOG_DEBUG, "cleaning up completed thread");
             pthread_join(threads[i].tid, NULL);
             threads[i].tid = 0;
         }
@@ -639,6 +650,12 @@ static void cleanup_context(spctx_t* ctx)
         ctx->sender = NULL;
     }
     
+    if(ctx->xforwardaddr)
+    {
+        free(ctx->xforwardaddr);
+        ctx->xforwardaddr = NULL;
+    }
+    
     ctx->logline[0] = 0;
 }
 
@@ -750,30 +767,38 @@ static int make_connections(spctx_t* ctx, int client)
         outaddr = &addr;
     }
     
-    /* No transparent proxy but check for loopback option */
-    else        
+    /* Check for loopback option */
+    else if(SANY_TYPE(*outaddr) == AF_INET && 
+            outaddr->s.in.sin_addr.s_addr == 0)
     {
-        if(SANY_TYPE(*outaddr) == AF_INET && 
-           outaddr->s.in.sin_addr.s_addr == 0)
-        {
-            /* Use the incoming IP as the default */
-            memcpy(&addr, &(g_state.outaddr), sizeof(addr));
-            memcpy(&(addr.s.in.sin_addr), &(peeraddr.s.in.sin_addr), sizeof(addr.s.in.sin_addr));
-            outaddr = &addr;
-        }
+        /* Use the incoming IP as the default */
+        memcpy(&addr, &(g_state.outaddr), sizeof(addr));
+        memcpy(&(addr.s.in.sin_addr), &(peeraddr.s.in.sin_addr), sizeof(addr.s.in.sin_addr));
+        outaddr = &addr;
+    }
 #ifdef HAVE_INET6        
-        else if(SANY_TYPE(*outaddr) == AF_INET6 && 
-                outaddr->s.in.in6.sin_addr.s_addr == 0)
-        {
-            /* Use the incoming IP as the default */
-            memcpy(&addr, &(g_state.outaddr), sizeof(addr));
-            memcpy(&(addr.s.in.sin6_addr), &(peeraddr.s.in.sin6_addr), sizeof(addr.s.in.sin6_addr));
-            outaddr = &addr;
-        }
+    /* IPv6 loopback? */
+    else if(SANY_TYPE(*outaddr) == AF_INET6 && 
+            outaddr->s.in.in6.sin_addr.s_addr == 0)
+    {
+        /* Use the incoming IP as the default */
+        memcpy(&addr, &(g_state.outaddr), sizeof(addr));
+        memcpy(&(addr.s.in.sin6_addr), &(peeraddr.s.in.sin6_addr), sizeof(addr.s.in.sin6_addr));
+        outaddr = &addr;
+    }
 #endif
-    }        
+
+    /* Not transparent proxy or loopback */
+    else 
+    {
+        /* Resolve any DNS name again */
+        if(sock_any_pton(g_state.outname, &addr, SANY_OPT_DEFPORT(25)) != -1)
+            memcpy(&(g_state.outaddr), &addr, sizeof(g_state.outaddr));
+        else
+            sp_messagex(ctx, LOG_WARNING, "couldn't resolve " CFG_OUTADDR ": %s", g_state.outname);
+    }
            
-    /* Reparse name if possible */
+    /* Reparse name if needed */
     if(outaddr != &(g_state.outaddr))
     {
         if(sock_any_ntop(outaddr, buf, MAXPATHLEN, 0) != -1)
@@ -874,7 +899,7 @@ static int smtp_passthru(spctx_t* ctx)
                     RETURN(-1);
                     
                 /* 
-                 * Now go into avcheck mode. This also handles the eventual
+                 * Now go into scan mode. This also handles the eventual
                  * sending of the data to the server, making the av check
                  * transparent
                  */
@@ -1041,7 +1066,8 @@ static int smtp_passthru(spctx_t* ctx)
                        is_first_word(p, ESMTP_CHUNK, KL(ESMTP_CHUNK)) ||
                        is_first_word(p, ESMTP_BINARY, KL(ESMTP_BINARY)) ||
                        is_first_word(p, ESMTP_CHECK, KL(ESMTP_CHECK)) ||
-                       is_first_word(p, ESMTP_XCLIENT, KL(ESMTP_XCLIENT)))
+                       is_first_word(p, ESMTP_XCLIENT, KL(ESMTP_XCLIENT)) ||
+                       is_first_word(p, ESMTP_XEXCH50, KL(ESMTP_XEXCH50)))
                     {
                         sp_messagex(ctx, LOG_DEBUG, "filtered ESMTP feature: %s", trim_space((char*)p));
                         
@@ -1059,7 +1085,7 @@ static int smtp_passthru(spctx_t* ctx)
                     }
                 }
 
-                /* MAIL FROM */
+                /* MAIL FROM (that the server accepted) */
                 if((r = check_first_word(C_LINE, FROM_CMD, KL(FROM_CMD), SMTP_DELIMS)) > 0)
                 {
                     t = parse_address(C_LINE + r);
@@ -1071,7 +1097,7 @@ static int smtp_passthru(spctx_t* ctx)
                         strcpy(ctx->sender, t);
                 }
                 
-                /* RCPT TO */
+                /* RCPT TO (that the server accepted) */
                 else if((r = check_first_word(C_LINE, TO_CMD, KL(TO_CMD), SMTP_DELIMS)) > 0)
                 {
                     t = parse_address(C_LINE + r);
@@ -1091,6 +1117,21 @@ static int smtp_passthru(spctx_t* ctx)
                         strcat(ctx->recipients, t);
                     }
                 }
+                            
+                /*
+                 * If the client sends an XFORWARD, and the server accepted it, 
+                 * we store address for use in our forked process environment 
+                 * variables (see sp_setup_forked). 
+                 */
+                else if(is_first_word(C_LINE, XFORWARD_CMD, KL(XFORWARD_CMD)))
+                {
+                    if((t = parse_xforward (C_LINE + KL(XFORWARD_CMD), "ADDR"))) 
+                    {
+                        ctx->xforwardaddr = (char*)reallocf(ctx->xforwardaddr, strlen(t) + 1);
+                        if(ctx->xforwardaddr)
+                            strcpy(ctx->xforwardaddr, t);
+                    }
+                }
                 
                 /* RSET */
                 else if(is_first_word(C_LINE, RSET_CMD, KL(RSET_CMD)))
@@ -1127,6 +1168,10 @@ static char* parse_address(char* line)
      * We parse out emails in the form of <blah@blah.com> 
      * as well as accept other addresses.
      */
+    
+    if(strncmp(line, "<>", 2) == 0)
+        return("<>");
+
     if(line[0] == '<')
     {
         if((t = strchr(line, '>')) != NULL)
@@ -1140,6 +1185,40 @@ static char* parse_address(char* line)
     return trim_end(line);
 }
 
+static char* parse_xforward(char* line, const char* part)
+{
+    char* t;
+    char* e;
+
+    t = strcasestr(line, part);
+    if(!t)
+        return NULL;
+
+    /* equals sign and surrounding */
+    t = trim_start(t + strlen(part));
+    if(*t != '=')
+        return NULL;
+    t = trim_start(t + 1);
+    if(!*t)
+        return NULL;
+    
+    /* Find the end of the thingy */
+    if(*t == '[')
+    {
+        t++;
+        e = strchr(t, ']');
+    }
+    else
+    {
+        e = t + strcspn(t, " \t");
+    }
+    
+    if(!e)
+        return NULL;
+    *e = 0;
+    return t;
+}
+
 static const char* get_successful_rsp(const char* line, int* cont)
 {
     /*
@@ -1473,20 +1552,26 @@ static int make_header(spctx_t* ctx, const char* format_str, char* header)
     return l >= MAX_HEADER_LENGTH ? MAX_HEADER_LENGTH - 1 : l; 
 }
 
-int sp_done_data(spctx_t* ctx)
+int sp_done_data(spctx_t* ctx, const char *headertmpl)
 {
     FILE* file = 0;
-    int had_header = 0;
     int ret = 0;
-    char line[SP_LINE_LENGTH];
-    char header[MAX_HEADER_LENGTH];
-    size_t header_len;
+    char *line;    
+    char header[MAX_HEADER_LENGTH] = "";
+    size_t header_len, line_len;
+    int header_prepend = 0;
+    ssize_t rc;
 
     ASSERT(ctx->cachename[0]);  /* Must still be around */
     ASSERT(!ctx->cachefile);    /* File must be closed */
     
     memset(header, 0, sizeof(header));
 
+    /* Alloc line buffer */
+    line_len = SP_LINE_LENGTH;
+    if((line = (char *)malloc(line_len)) == NULL)
+        RETURN(-1);
+
     /* Open the file */
     file = fopen(ctx->cachename, "r");
     if(file == NULL)
@@ -1515,20 +1600,25 @@ int sp_done_data(spctx_t* ctx)
 
     sp_messagex(ctx, LOG_DEBUG, "sending from cache file: %s", ctx->cachename);
     
-    if(g_state.header)
-        header_len = make_header(ctx, g_state.header, header);
+    if(headertmpl)
+    {
+        header_len = make_header(ctx, headertmpl, header);
+        if(is_first_word(RCVD_HEADER, header, KL(RCVD_HEADER)))
+            header_prepend = 1;
+    }
+
 
     /* If we have to prepend the header, do it */
-    if(header[0] && g_state.header_prepend)
+    if(header[0] != '\0' && header_prepend)
     {
-	    if(spio_write_data_raw(ctx, &(ctx->server), (char*)header, header_len) == -1 ||
-	       spio_write_data_raw(ctx, &(ctx->server), CRLF, KL(CRLF)) == -1)
-	        RETURN(-1);
-	    had_header = 1;
+        if(spio_write_data_raw(ctx, &(ctx->server), (unsigned char*)header, header_len) == -1 ||
+           spio_write_data_raw(ctx, &(ctx->server), (unsigned char*)CRLF, KL(CRLF)) == -1)
+            RETURN(-1);
+        header[0] = '\0';
     }
 
     /* Transfer actual file data */    
-    while(fgets(line, SP_LINE_LENGTH, file) != NULL)
+    while((rc = getline(&line, &line_len, file)) != -1)
     {
         /* 
          * If the line is <CRLF>.<CRLF> we need to change it so that 
@@ -1539,7 +1629,7 @@ int sp_done_data(spctx_t* ctx)
         if(strcmp(line, "." CRLF) == 0)
             strncpy(line, ". " CRLF, SP_LINE_LENGTH);
       
-        if(header[0] && !had_header)
+        if(header[0] != '\0')
         {
             /* 
              * The first blank line we see means the headers are done.
@@ -1547,15 +1637,14 @@ int sp_done_data(spctx_t* ctx)
              */
             if(is_blank_line(line))
             {
-                if(spio_write_data_raw(ctx, &(ctx->server), (char*)header, header_len) == -1 ||
-                   spio_write_data_raw(ctx, &(ctx->server), CRLF, KL(CRLF)) == -1)
+                if(spio_write_data_raw(ctx, &(ctx->server), (unsigned char*)header, header_len) == -1 ||
+                   spio_write_data_raw(ctx, &(ctx->server), (unsigned char*)CRLF, KL(CRLF)) == -1)
                     RETURN(-1);
-            
-                had_header = 1;
+                header[0] = '\0';
             }
         }
         
-        if(spio_write_data_raw(ctx, &(ctx->server), line, strlen(line)) == -1)
+        if(spio_write_data_raw(ctx, &(ctx->server), (unsigned char*)line, rc) == -1)
             RETURN(-1);
     }
     
@@ -1580,6 +1669,8 @@ int sp_done_data(spctx_t* ctx)
         
 cleanup:
     
+	if(line)
+		free(line); 
     if(file)
         fclose(file); /* read-only so no error check */
     
@@ -1678,6 +1769,9 @@ void sp_setup_forked(spctx_t* ctx, int file)
     if(spio_valid(&(ctx->client)))
         setenv("CLIENT", ctx->client.peername, 1);
     
+    if(ctx->xforwardaddr)
+        setenv("REMOTE", ctx->xforwardaddr, 1);
+    
     if(spio_valid(&(ctx->server)))
         setenv("SERVER", ctx->server.peername, 1);
     
@@ -1690,13 +1784,12 @@ void sp_setup_forked(spctx_t* ctx, int file)
  */
  
 const char kMsgDelimiter[] = ": ";
-#define MAX_MSGLEN  256
+#define MAX_MSGLEN  1024
 
 static void vmessage(spctx_t* ctx, int level, int err, 
                      const char* msg, va_list ap)
 {
-    size_t len;
-    char* m;  
+    char buf[MAX_MSGLEN];
     int e = errno;
 
     if(g_state.daemonized)
@@ -1712,35 +1805,30 @@ static void vmessage(spctx_t* ctx, int level, int err,
        
     ASSERT(msg);
 
-    len = strlen(msg) + 20 + MAX_MSGLEN;
-    m = (char*)alloca(len);
-
-    if(m)
+    if(ctx)
+        snprintf(buf, MAX_MSGLEN, "%06X: %s%s", ctx->id, msg, err ? ": " : "");
+    else
+        snprintf(buf, MAX_MSGLEN, "%s%s", msg, err ? ": " : "");
+            
+    if(err)
     {
-        if(ctx)
-            snprintf(m, len, "%06X: %s%s", ctx->id, msg, err ? ": " : "");
-        else
-            snprintf(m, len, "%s%s", msg, err ? ": " : "");
+        /* strerror_r doesn't want to work for us for some reason
+        len = strlen(buf);
+        strerror_r(e, buf + len, MAX_MSGLEN - len); */
             
-        if(err)
-        {
-            /* strerror_r doesn't want to work for us for some reason
-            strerror_r(e, m + strlen(m), MAX_MSGLEN); */
-            
-            sp_lock();
-                strncat(m, strerror(e), len);
-            sp_unlock();
-        }
-            
-        m[len - 1] = 0;
-        msg = m;
-    }    
+        sp_lock();
+            strncat(buf, strerror(e), MAX_MSGLEN);
+        sp_unlock();
+    }
+
+    /* As a precaution */
+    buf[MAX_MSGLEN - 1] = 0;
   
     /* Either to syslog or stderr */
     if(g_state.daemonized)
-        vsyslog(level, msg, ap);
+        vsyslog(level, buf, ap);
     else
-        vwarnx(msg, ap);  
+        vwarnx(buf, ap);  
 }
 
 void sp_messagex(spctx_t* ctx, int level, const char* msg, ...)
@@ -1903,16 +1991,6 @@ int sp_parse_option(const char* name, const char* value)
         ret = 1;
     }
 
-    else if(strcasecmp(CFG_HEADER, name) == 0)
-    {
-        g_state.header = trim_start(value);
-	    if(strlen(g_state.header) == 0)
-	        g_state.header = NULL;
-	    else if(is_first_word(RCVD_HEADER, g_state.header, KL(RCVD_HEADER)))
-	        g_state.header_prepend = 1;
-	    ret = 1;
-    }
-    
     /* Always pass through to program */    
     if(cb_parse_option(name, value) == 1)
         ret = 1;
